#!/usr/bin/perl -w

# badger.pl (c) 2000 Chris Snell <chris@blueaspen.com>
# badger.pl reads badger.conf and makes SNMP calls
# and exec's programs to gather data, which is then
# fed into a RRD file using the RRD Perl module.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


use strict;
use POSIX;
use RRDs;
use Net::SNMP ();


use Time::HiRes qw( usleep gettimeofday );


##------------------------------------------------------------
## CONFIGURATION
##------------------------------------------------------------
my $CONF          = '/etc/bronc/badger.conf';
my $COMMUNITY_MAP = '/etc/bronc/community_map';
my $RRD_PATH      = '/etc/bronc/rrd_files';
my $PORT          = 161;



##------------------------------------------------------------
## Misc globals
##------------------------------------------------------------
my $caught_hup       = 0;
my $time_to_die      = 0;
my $start_time       = time;
my (%jobs,%comm_map) = ();



##------------------------------------------------------------
## Signal Handlers
##------------------------------------------------------------
$SIG{INT}  = $SIG{TERM} = \&signal_handler;
$SIG{HUP}  = \&signal_hiccup;
$SIG{ALRM} = \&launcher;



##------------------------------------------------------------
## Fork into background
##------------------------------------------------------------
my $pid = fork;
exit if $pid;
die "Could not fork: $!" unless defined ($pid);
POSIX::setsid() || die "Cannot start a new session: $!\n";



##------------------------------------------------------------
## Read configuration files
##------------------------------------------------------------
read_communities();
read_conf();



##------------------------------------------------------------
## Open our debug log file for output 
##    (hrm, what about syslog?)
##------------------------------------------------------------
open(DEBUG2, ">/var/log/bronc/snmp-debug.log") or
    warn "Cannot open snmp-debug.log\n";


##------------------------------------------------------------
## Until we receive a signal of some type, we'll loop here
## through all the objects to collect data for.
##------------------------------------------------------------
until ($time_to_die) {


    ##--------------------------------------------------------
    ## Do the data collection until we receive a SIGHUP.  If
    ## we receive a SIGHUP, we'll restart the loop with the
    ## new data in the config files.
    ##--------------------------------------------------------    
    until ($caught_hup || $time_to_die) {


        usleep(2000);
        
        ##----------------------------------------------------
        ## Get the current time and subtract the start time to
        ## get the elapsed seconds since we started.
        ##----------------------------------------------------
#        my $this_time = gettimeofday;
        my $this_time = time;
        my $span = $this_time - $start_time;
        
        
        ##----------------------------------------------------
        ## Now loop through all of the jobs
        ##----------------------------------------------------
        foreach my $a_job (keys %jobs) {
#           warn "A_JOB: $a_job\n";
#           warn "THIS_TIME: $this_time\n";
#           warn "START_TIME: $start_time\n";
#           warn "SPAN: $span\n";
#           warn "INTERVAL: $jobs{$a_job}{'interval'}\n";
            
            
            ##------------------------------------------------
            ## If we're not on a span multiple of the query
            ## interval AND we didn't just collect the data for
            ## this period, then let's launch the collector.
            ##------------------------------------------------
            if ( (! ($span % $jobs{$a_job}{'interval'})) && 
                 ($jobs{$a_job}{'last_launched'} != $this_time) ) 
            {
#		warn "LAUNCH: $jobs{$a_job}{'hostname'} @{$jobs{$a_job}{'oids'}} $a_job \n";

		if ($jobs{$a_job}{'hostname'} =~ m/\|/) {
			my $return_val = exec_launcher("${RRD_PATH}/$jobs{$a_job}{'rrd_file'}",$jobs{$a_job}{'oids'});
		} else {
	                my $return_val = launcher($jobs{$a_job}{'hostname'},
                                          $jobs{$a_job}{'community'},
                                          $jobs{$a_job}{'oids'}, 
                                          $a_job);
		}
                
                $jobs{$a_job}{'last_launched'} = $this_time;
            }
            
            
            ##------------------------------------------------
            ## If the current time minus the last time we 
            ## collected data is greater than the interval, 
            ## we'll also launch.
            ##
            ## NOTE: This differs from the launcher above as 
            ##       this will try to collect data from 
            ##       machines that haven't been queried in 
            ##       previous iterations.
            ##------------------------------------------------
            if ( ($this_time - $jobs{$a_job}{'last_launched'}) > $jobs{$a_job}{'interval'} ) 
            {
		warn "LAUNCH: $jobs{$a_job}{'hostname'} @{$jobs{$a_job}{'oids'}} $a_job \n";

                if ($jobs{$a_job}{'hostname'} =~ m/\|/) {
			my $return_val = exec_launcher("${RRD_PATH}/$jobs{$a_job}{'rrd_file'}",$jobs{$a_job}{'oids'});

                } else {
	                my $return_val = launcher($jobs{$a_job}{'hostname'},
                                          $jobs{$a_job}{'community'},
                                          $jobs{$a_job}{'oids'}, 
                                          $a_job);
		}
                
                $jobs{$a_job}{'last_launched'} = $this_time;
            }
        }
    }
    
    $caught_hup = 0;
}



##------------------------------------------------------------
## Handle signals by closing the DEBUG file and setting the
## time_to_do variable true.
##------------------------------------------------------------
sub signal_handler 
{
    close(DEBUG2);
    $time_to_die = 1;
}


sub signal_hiccup 
{
    read_communities();
    read_conf();


    $caught_hup = 1;
}

##------------------------------------------------------------
## exec_launcher
##
## This launches local executables and stores the result in
## an RRD.
##------------------------------------------------------------
sub exec_launcher {
    my $argument;
    my $output;
    my $command_line;
    my ($rrd_file, $prog_array_ref) = @_;

    ##--------------------------------------------------------
    ## Build our command line in preparation for the exec call
    ##--------------------------------------------------------
    foreach $argument (@{$prog_array_ref}) {
	$command_line .= $argument . " ";
	chomp($command_line);
    }
    #    print "COMMANDLINE: $command_line\n";

    ##--------------------------------------------------------
    ## Execute the commandline and snarf the returned value(s)
    ##--------------------------------------------------------
    my $pid = open(PROGGIE, "$command_line |") 
          or warn "$command_line cannot be executed. \n Check file and permissions for validity\n.";
   
    while (<PROGGIE>) {
      $output = $_;
      #  print "OUTPUT: $output\n";
    }
    close(PROGGIE);
   
    my $right_now = time;

    $output =~ s/^\s+//;            # No leading white
    $output =~ s/\s+$//;            # No trailing white


    ##--------------------------------------------------------
    ## Finally, update the RRD.
    ##--------------------------------------------------------
    RRDs::update ("$rrd_file", "${right_now}:$output");
    # print "RRD_UPDATE $rrd_file ${right_now}:$output \n";


    ##--------------------------------------------------------
    ## And die if we got an error
    ##--------------------------------------------------------
    my $ERR=RRDs::error;
    warn "ERROR while updating $rrd_file [ $right_now
: $output ]: $ERR\n" if $ERR;


    ##--------------------------------------------------------
    ## Else, return the values.
    ##--------------------------------------------------------
    return($output);

}

##------------------------------------------------------------
## launcher
##
## All the dirty work is done here.
##------------------------------------------------------------
sub launcher {
    my ($l_snmphost, $l_community, $l_oid, $a_job) = @_;


    my @ret_vals     = ();
    my $snmp_vals    = "";
    my $oids_to_get  = "";
    my $returned_oid = "";
    my $response     = "";


#       print "snmphost: $l_snmphost oid: @{$l_oid}  comm: $l_community\n";


    ##--------------------------------------------------------
    ## Create a new SNMP session to the host
    ##--------------------------------------------------------
    my ($session, $error) = Net::SNMP->session(-hostname  => $l_snmphost,
                                               -community => $l_community);


    ##--------------------------------------------------------
    ## If we can't, log the error and return
    ##--------------------------------------------------------
    if ( !defined($session) ) {
        printf("ERROR: %s.\n", $error);
        print DEBUG2 "ERROR: $error\n";
        return;
    }


    ##--------------------------------------------------------
    ## Fetch the OID value and log error, return if not
    ## defined.
    ##--------------------------------------------------------
    if (!defined($response = $session->get_request(@{$l_oid}))) {
        printf("ERROR: %s.\n", $session->error);
        $session->close;
        print DEBUG2 "ERROR: $error\n";
        return;
    }


    ##--------------------------------------------------------
    ## Fetch the current epoch time
    ##--------------------------------------------------------
    my $right_now = time;


    ##--------------------------------------------------------
    ## Walk through the returned values and create a scalar
    ## This is messy and should be cleaned up.
    ##--------------------------------------------------------
    foreach $returned_oid (keys %{$response}) {
        $snmp_vals = "$snmp_vals $response->{$returned_oid}";
    }


    ##--------------------------------------------------------
    ## And globall replace all spaces with a colon because 
    ## the RRD Perl module likes them this way.
    ##--------------------------------------------------------
    $snmp_vals =~ s/ /:/g;


    ##--------------------------------------------------------
    ## Close the SNMP session
    ##--------------------------------------------------------
    $session->close;


    ##--------------------------------------------------------
    ## Finally, update the RRD.
    ##--------------------------------------------------------
  RRDs::update ("${RRD_PATH}/$jobs{$a_job}{'rrd_file'}", "${right_now}$snmp_vals");


    ##--------------------------------------------------------
    ## And die if we got an error
    ##--------------------------------------------------------
    my $ERR=RRDs::error;
    warn "ERROR while updating ${RRD_PATH}/$jobs{$a_job}{'rrd_file'} [ $right_now:$snmp_vals ]: $ERR\n" if $ERR;


    ##--------------------------------------------------------
    ## Else, return the values.
    ##--------------------------------------------------------
    return($snmp_vals);
}


##------------------------------------------------------------
## read_communities
##
## Read the community map
##------------------------------------------------------------
sub read_communities
{
    open(CMAP,"<$COMMUNITY_MAP") || die "Cannot open $COMMUNITY_MAP.\n";
    
    while(<CMAP>) {
        chomp;               # No newline
        s/#.*//;             # No comments
        s/^\s+//;            # No leading white
        s/\s+$//;            # No trailing white
        next unless length;  # Anything left?
        
#       warn "LINE: $_\n";
        my @line = split /\s+/;
        $comm_map{$line[0]} = $line[1];
    }
    close(CMAP);
}


##------------------------------------------------------------
## read_conf
##
## Read the badger.conf file
##------------------------------------------------------------
sub read_conf
{
    open(CFG,"<$CONF") || die "cannot open $CONF\n";
    
    my $a_count = 0;
    
    my $keycount = 0;
    while (<CFG>) {
        chomp;               # No newline
        s/#.*//;             # No comments
        s/^\s+//;            # No leading white
        s/\s+$//;            # No trailing white
        next unless length;  # Anything left?
        
        my @conf_line = split /\s+/;
        # print "keycount: $keycount\n";
        $jobs{$keycount}{'interval'}  = $conf_line[0];
        $jobs{$keycount}{'rrd_file'}  = $conf_line[1];
        $jobs{$keycount}{'hostname'}  = $conf_line[2];
        $jobs{$keycount}{'last_launched'}  = 0;
        my $temp_hostname = $jobs{$keycount}{'hostname'};
        $jobs{$keycount}{'community'} = $comm_map{$temp_hostname};
        my $array_size = scalar @conf_line;
        $a_count = 0;
        my @oid_holder = ();
        while (($a_count + 3) < $array_size) {
            push(@oid_holder,$conf_line[3 + $a_count]);
            $a_count++;
        }
        $jobs{$keycount}{'oids'}      = [ @oid_holder ];
        $keycount++;
    }
}


##------------------------------------------------------------
## eof
##------------------------------------------------------------
